/*
******************************************************************
*    Dateiname: ft1.cpp   Datum: 25.10. 2003     Autor: Pitri    *
******************************************************************
*/

/* Includes */
/* --------*/
#include "stdlib.h"
#include "stdio.h"
#include "math.h"

int main
/******/
/*
Multiplikation mit der Fourier Transformation.

Folgende Teilschritte werden ausgefhrt.
1. Einlesend der Multiplikanden
2. Ermitteln von n
3. Alle Ziffern separat speichern.
4. Alle n-ten Einheitswurzeln berechnen.
5. Diskrete Fouriertransformation mit jedem Multiplikanden durchfhren. (DFT des Vektors)
6. Multiplikation der transformierten Vektoren (Multiplikation von jeweils zwei komplexen Zahlen)
7. Inverse Fouriertransformation mit dem Ergebnis der DFT durchfhren.
*/
(
  int argc,
  char *argv[]
)
{
  // Funktionen
  int get_len_long(unsigned long);
  void split_to_numerals(unsigned long,int*);
  void get_einheitswurzel(int n,double*wr,double*wi);
  void get_dft_vektor(int n,int*numerals,int len_numerals,double*wr,double*wi,double*dft_r,double*dft_i);
  void mul_dft(int n,double*dft_r1,double*dft_i1,double*dft_r2,double*dft_i2,double*res_dft_r,double*res_dft_i);
  void inverse_dft_vektor(int n,double*wr,double*wi,double*dft_r,double*dft_i,double*inf_dft_r,double*inf_dft_i,double*res_polynom);
  void decode_polynom(double*polynom,int polynom_len);
  void round(double*polynom,int polynom_len);

   // Variablen
  unsigned long m1;
  unsigned long m2;
  int len_m1;
  int len_m2;
  int n;
  int i1;
  int numerals_m1[200];
  int numerals_m2[200];
  double wr[200];
  double wi[200];
  double dft_r1[200];
  double dft_i1[200];
  double dft_r2[200];
  double dft_i2[200];
  double res_dft_r[200];
  double res_dft_i[200];
  double inf_dft_r[200];
  double inf_dft_i[200];
  double res_polynom[200];

  printf("Multiplikand  m1 eingeben:");
  scanf("%lu",&m1);

  printf("Multiplikator m2 eingeben:");
  scanf("%lu",&m2);

  len_m1=get_len_long(m1);
  len_m2=get_len_long(m2);

  // folgende n zuweisung bin ich nicht sicher ob n nicht zwingend eine 2er Potenz sein muss....??
  if(len_m1>len_m2)
  {
    n=2*len_m1;
  }
  else
  {
    n=2*len_m2;
  }

  printf("\n");
  printf("laenge ist:%d\n",len_m1);
  printf("laenge ist:%d\n",len_m2);

  split_to_numerals(m1,numerals_m1);
  split_to_numerals(m2,numerals_m2);

  get_einheitswurzel(n,wr,wi);

  printf("\n");
  printf("%d.Einheitswurzel:\n",n);
  for(i1=0;i1<n;i1++)
  {
    printf("w%d= ( %+.10f , %+.10fi )\n",i1,wr[i1],wi[i1]);
  }

  get_dft_vektor(n,numerals_m1,len_m1,wr,wi,dft_r1,dft_i1);
  get_dft_vektor(n,numerals_m2,len_m2,wr,wi,dft_r2,dft_i2);

  printf("\n");
  printf("DFT des Vektors %lu:\n",m1);
  for(i1=0;i1<n;i1++)
  {
    printf("[%d]= ( %+.10f , %+.10fi )\n",i1,dft_r1[i1],dft_i1[i1]);
  }
  printf("\n");
  printf("DFT des Vektors %lu:\n",m2);
  for(i1=0;i1<n;i1++)
  {
    printf("[%d]= ( %+.10f , %+.10fi )\n",i1,dft_r2[i1],dft_i2[i1]);
  }

  mul_dft(n,dft_r1,dft_i1,dft_r2,dft_i2,res_dft_r,res_dft_i);

  printf("\n");
  printf("DFT Multiplikation mit den Vektoren %lu und %lu:\n",m1,m2);
  for(i1=0;i1<n;i1++)
  {
    printf("[%d]= ( %+.10f , %+.10fi )\n",i1,res_dft_r[i1],res_dft_i[i1]);
  }

  inverse_dft_vektor(n,wr,wi,res_dft_r,res_dft_i,inf_dft_r,inf_dft_i,res_polynom);

  printf("\n");
  printf("Inverse DFT\n");
  for(i1=0;i1<n;i1++)
  {
    printf("[%d]= ( %+.10f , %+.10fi )\n",i1,inf_dft_r[i1],inf_dft_i[i1]);
  }

  printf("\n");
  printf("Ergebnis Polynom\n");
  for(i1=0;i1<n;i1++)
  {
    printf("[%d]=%+.10f\n",i1,res_polynom[i1]);
  }

  round(res_polynom,n);

  printf("\n");
  printf("Ergebnis Polynom gerundet\n");
  for(i1=0;i1<n;i1++)
  {
    printf("[%d]=%+.10f\n",i1,res_polynom[i1]);
  }

  decode_polynom(res_polynom,n);
  
  printf("\n");
  printf("Polynom decodiert in eine Dezimalzahl\n");
  for(i1=0;i1<n;i1++)
  {
    printf("[%d]=%+.10f\n",i1,res_polynom[i1]);
  }

  return(0);
}


int get_len_long
/**************/
(
  unsigned long z1
)
/*
  Beschreibung:
  Es wird die Anzahl Ziffern der uebergebenen Longzahl ermittelt.
*/
{
  unsigned long i1=10;
  int len=1;

  while(z1/i1)
  {
    i1*=10;
    len++;
  }
  return(len);
}


void split_to_numerals
/********************/
(
  unsigned long z1,
  int *numerals
)
/*
  Beschreibung:
  Die einzelnen Ziffern der uebergebenen Longzahl werden extrahiert.
  In numerals[0] steht die niederwertigste Ziffer.
*/
{
  unsigned long i1=10;
  int len=1;

  while(z1/i1)
  {
    i1*=10;
    len++;
  }

  i1/=10;

  while(len>0)
  {
    numerals[len-1]=(int)(z1/i1);
    z1-=numerals[len-1]*i1;
    i1/=10;
    len--;
  }
}



void get_einheitswurzel
/*********************/
(
  int n,
  double *wr,
  double *wi
)
/*
  Beschreibung:
  Es werden alle n Einheitswurzeln berechnet und in wr und wi eingetragen.
  In wr steht der Realanteil, in wi der Imaginaeranteil.
*/
{
  int i1;
  double delta_phi;
  double phi;

  delta_phi=(double)2*3.14159/n;
  phi=0;

  for(i1=0;i1<n;i1++)
  {
    wr[i1]=cos(phi);
    wi[i1]=sin(phi);
    phi+=delta_phi;
  }
}


void get_dft_vektor
/*****************/
(
  int n,
  int *numerals,
  int len_numerals,
  double *wr,
  double *wi,
  double *dft_r,
  double *dft_i
)
/*
  Beschreibung:
  Es werden dft_r und dft_i berechnet.
  Alle anderen uebergabeparameter sind bei aufruf dieser Funktion bereits abgefuellt.
  In dft_r steht der Realanteil, in dft_i der Imaginaeranteil.
*/
{
  int i1;
  int i2;
  int i3;
  int i4;
  
  for(i1=0;i1<n;i1++)
  {
    dft_r[i1]=0.0;
    dft_i[i1]=0.0;
  }

  for(i2=0;i2<len_numerals;i2++)
  {
    dft_r[0]+=(double)numerals[i2]; // *1 weil der r-Anteil immer 1 ist.
    //dft_i[0]+=(double)numerals[i2]*0;  weglassen weil der i-Anteil immer 0 ist.
  }

  i3=1;
  for(i1=1;i1<n;i1++)
  {
    for(i2=0;i2<len_numerals;i2++)
    {
      i4=(i2*i3)%n;
      dft_r[i1]+=numerals[i2]*(wr[i4]);
      dft_i[i1]+=numerals[i2]*(wi[i4]);
    }
    i3++;
  }
}


void mul_dft
/**********/
(
  int n,
  double *dft_r1,
  double *dft_i1,
  double *dft_r2,
  double *dft_i2,
  double *res_dft_r,
  double *res_dft_i
)
/*
  Beschreibung:
  Es werden dft_r1 und dft_i1 multipliziert mit dft_r2 und dft_i2.
  Alle dft sind Arrays der groesse n. Jeder dft Arrayplatz ist eine komplexe Zahl.
  Es werden also n mal komplexe Zahlen multipliziert und in res_dft_r,res_dft_i eingetragen.
  _r steht fuer Realanteil, _i steht fuer Imaginaeranteil.
*/
{
  int i1;
  
  // fuer die Multiplikation von komplexen Zahlen gilt: (a,b)*(c,d)=(a*c-b*d,a*d+b*c)
  for(i1=0;i1<n;i1++)
  {
    res_dft_r[i1]=(dft_r1[i1]*dft_r2[i1])-(dft_i1[i1]*dft_i2[i1]);
    res_dft_i[i1]=(dft_r1[i1]*dft_i2[i1])+(dft_i1[i1]*dft_r2[i1]);
  }
}


void inverse_dft_vektor
/*********************/
(
  int n,
  double *wr,
  double *wi,
  double *dft_r,
  double *dft_i,
  double *inf_dft_r,
  double *inf_dft_i,
  double *res_polynom
)
/*
  Beschreibung:
  Es werden aus dft_r und dft_i das Ergebnis berechnet. Das Ergebnis wird in inf_dft_r
  und inf_dft_i eingetragen.
  Alle anderen uebergabeparameter sind bei aufruf dieser Funktion bereits abgefuellt.
  _r steht fuer Realanteil, _i fuer Imaginaeranteil.
*/
{
  int i1;
  int i2;
  int i3;
  int i4;
  
  for(i1=0;i1<n;i1++)
  {
    inf_dft_r[i1]=0.0;
    inf_dft_i[i1]=0.0;
  }

  for(i2=0;i2<n;i2++)
  {
    inf_dft_r[0]+=dft_r[i2]; // *1 weil der r-Anteil immer 1 ist.
    //inf_dft_i[0]+=dft_i[i2]*0;  weglassen weil der i-Anteil immer 0 ist.
  }

  i3=1;
  for(i1=1;i1<n;i1++)
  {
    for(i2=0;i2<n;i2++)
    {
      i4=(i2*i3)%n;
      inf_dft_r[i1]+=dft_r[i2]*(wr[i4]);
      inf_dft_i[i1]+=dft_i[i2]*(wi[i4]);
    }
    i3++;
  }

  for(i1=0;i1<n;i1++)
  {
    res_polynom[i1]=(inf_dft_r[i1]+inf_dft_i[i1])/n;
  }

  for(i1=0;i1<n;i1++)
  {
    inf_dft_r[i1]/=n;
    inf_dft_i[i1]/=n;
  }
}


void decode_polynom
/*****************/
(
  double *polynom,
  int polynom_len
)
/*
  Beschreibung:
  Ein Polynom wird in eine Dezimale Zahl ueberfuehrt.
  Achtung das Polynom muss bereits gerundet sein.
  In polynom[0] steht die niederwertigste Ziffer.
  Die Dezimale Zahl wird wieder in das Array polynom eingetragen.
*/
{
  int i1;
  long tmp1=0;
  long polynom_tmp;

  for(i1=0;i1<polynom_len;i1++)
  {
    polynom_tmp=(long)polynom[i1]+tmp1;
    if(polynom_tmp>9)
    {
      tmp1=polynom_tmp/10;
      polynom_tmp%=10;
      polynom[i1]=(double)polynom_tmp;
    }
    else if(polynom_tmp<0)
    {
      tmp1=(polynom_tmp/10)-1;
      polynom_tmp=tmp1*(-10)+polynom_tmp;
      polynom[i1]=(double)polynom_tmp;
    }
    else
    {
      tmp1=0;
      polynom[i1]=(double)polynom_tmp;
    }
  }
}



void round
/********/
(
  double *polynom,
  int polynom_len
)
/*
Beschreibung:
Die in polynom uebergebenen double Werte werden auf die naechsten ganzzahligen Werte gerundet.Das 
ganzzahlige Ergebnis wird wiederum in polynom eingetragen.
*/
{
  long   temp_i;
  double temp_f;
  int i1;

  for(i1=0;i1<polynom_len;i1++)
  {
    temp_i=(long)polynom[i1];
    temp_f=polynom[i1]-temp_i;

    if(polynom[i1]>0.0)
    {
      if (temp_f<0.5)
      { 
        /* abrunden */
        polynom[i1]=temp_i;
      }
      else
      {
        /* aufrunden */
        polynom[i1]=temp_i+1;
      }
    }
    else
    {
      if (temp_f>(-0.5)) // >-0.5 ist bsp. -0.4999
      { 
        /* abrunden */
        polynom[i1]=temp_i;
      }
      else
      {
        /* aufrunden */
        polynom[i1]=temp_i-1;
      }
    }
  }
}
